/*
 Copyright (c) 2013 aizo ag, Zurich, Switzerland

 Author: Christain Hitz <christian.hitz@aizo.com>

 This file is part of dS Assistant.

 dS-Assistant is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 dS-Assistant is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

 NOTE: all APIs, especially the JSON API, offered by this application
 are subject to change without further notice.
 */

#include "comm-channel.h"
#include "exceptions.h"
#include "ds-assistant.h"
#include <pthread.h>
#include <libcommchannel/commchannel.h>
#include <libcommchannel/exceptions.h>
#include "log.h"
#include "gettext_helper.h"
#include "messages/messaging.pb.h"
#include "messages/device-scene-table-update.pb.h"
#include "tasks.h"
#include "config_manager.h"

using dSA::Exception;

namespace dSA {

CommChannel* CommChannel::m_instance = NULL;

CommChannel* CommChannel::createInstance()
{
    if (m_instance == NULL)
    {
        m_instance = new CommChannel();
        if (m_instance == NULL)
        {
            throw _Exception("could not create CommChannel class");
        }
    }

    return m_instance;
}

CommChannel* CommChannel::getInstance()
{
    if (m_instance == NULL)
    {
        throw _Exception("CommChannel not initialized");
    }

    return m_instance;
}

CommChannel::CommChannel()
        : m_shutdown_flag(false),
          m_running(false),
          m_thread(0)
{
    int ret;

    ret = pthread_mutex_init(&m_mutex, NULL);
    if (ret != 0)
    {
        throw _Exception("failed to initialize message list mutex: " +
                std::string(strerror(errno)));
    }

    ret = pthread_cond_init(&m_condition, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_mutex);
        throw _Exception("failed to initialize message handling wakeup condition: " +
                std::string(strerror(errno)));
    }

    int dSAPort = ConfigManager::getInstance()->getIntOption(CFG_COMM_CHANNEL_DSA_PORT);
    m_pServer = new CC::CommunicationChannelServer(dSAPort);
    if (!m_pServer)
    {
        pthread_mutex_destroy(&m_mutex);
        pthread_cond_destroy(&m_condition);
        throw _Exception("failed to initialize CommunicationChannelServer");
    }

    m_pServer->addCallback(this);
}

void CommChannel::shutdown()
{
    m_shutdown_flag = true;

    lockMessageList();
    pthread_cond_signal(&m_condition);
    unlockMessageList();
    if (m_thread)
    {
        pthread_join(m_thread, NULL);
        m_thread = 0;
    }
}

CommChannel::~CommChannel()
{
    shutdown();

    if (m_pServer)
    {
        m_pServer->removeCallback(this);
        delete m_pServer;
    }

    m_instance = NULL;

    pthread_mutex_destroy(&m_mutex);
    pthread_cond_destroy(&m_condition);
}

void CommChannel::lockMessageList()
{
    int ret = pthread_mutex_lock(&m_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to lock mutex: " +
                std::string(strerror(errno)));
    }
}

void CommChannel::unlockMessageList()
{
    int ret = pthread_mutex_unlock(&m_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to unlock mutex: " +
                std::string(strerror(errno)));
    }
}

void CommChannel::messageReceived(boost::shared_ptr<CC::CommunicationData> comm)
{
    lockMessageList();
    m_datalist.push_back(comm);
    pthread_cond_signal(&m_condition);
    unlockMessageList();
}

void CommChannel::run()
{
    m_running = true;

    int ret = pthread_create(&m_thread, NULL, CommChannel::staticThreadProc, this);
    if (ret != 0)
    {
        m_running = false;
        throw _Exception("Could not start task thread");
    }
}

std::string CommChannel::sendMessage(const std::string& message)
{
    int dSSPort = ConfigManager::getInstance()->getIntOption(CFG_COMM_CHANNEL_DSS_PORT);
    CC::CommunicationChannelClient cc(dSSPort);

    try
    {
        boost::shared_ptr<CC::CommunicationData> c = cc.sendData((unsigned char*) message.c_str(), message.length());
        const char *data = (const char *) c->getDataPtr();
        if (data != NULL)
        {
            return std::string(data);
        }
        else
        {
            FILE_LOG(logERROR) << "[CommChannel] sending message failed: no response";
        }
    }
    catch (CC::Exception& ex)
    {
        FILE_LOG(logERROR) << "[CommChannel] sending message failed: " << ex.getMessage();
    }
    return std::string();
}

bool CommChannel::sendSceneLocks(const std::vector<int>& scenes)
{
    dsmsg::Message message;
    message.set_type(dsmsg::DEVICE_SCENE_TABLE_UPDATE_STATE);
    dsmsg::DeviceSceneTableUpdateState *updateState = message.MutableExtension(dsmsg::device_scene_table_update_state);
    for (std::vector<int>::const_iterator sceneIt = scenes.begin(); sceneIt != scenes.end(); ++sceneIt)
    {
        updateState->add_locked_scene(*sceneIt);
    }

    std::string response = sendMessage(message.SerializeAsString());
    if (!response.empty())
    {
        message.Clear();
        message.ParseFromString(response);
        if (message.generic_res().result() != 0)
        {
            FILE_LOG(logERROR) << "[CommChannel] response: " << message.generic_res().description();
            return false;
        }
    }
    return true;
}

void *CommChannel::staticThreadProc(void *arg)
{
    CommChannel *instance = (CommChannel *) arg;
    instance->taskThread();
    pthread_exit(NULL);
    return NULL;
}

void CommChannel::taskThread()
{
    try
    {
        lockMessageList();
        while (!m_shutdown_flag)
        {
            if (m_datalist.empty())
            {
                pthread_cond_wait(&m_condition, &m_mutex);
                continue;
            }

            boost::shared_ptr<CC::CommunicationData> c = m_datalist.front();
            m_datalist.pop_front();

            unlockMessageList();

            const unsigned char *data = c->getDataPtr();
            if (data != NULL)
            {
                dsmsg::Message response;
                response.set_type(dsmsg::GENERIC_RESPONSE);
                dsmsg::ResultCode result = dsmsg::RESULT_OK;
                std::string resultDescription = "OK";

                dsmsg::Message received_message;
                std::string dataString((const char *) data);
                received_message.ParseFromString(dataString);
                switch (received_message.type())
                {
                    case dsmsg::DEVICE_SCENE_TABLE_UPDATE_START:
                    {
                        dsuid_t devid;
                        dsuid_t dsmid;
                        boost::shared_ptr<const dSLongTaskInfo> task = dS_Assistant::getInstance()->getTaskInfoByType(
                                TestDeviceUpdateTask);
                        if (task == NULL)
                        {
                            SetNullDsuid(devid);
                            SetNullDsuid(dsmid);
                            boost::shared_ptr<DeviceUpdateTask> task(new DeviceUpdateTask(dsmid, devid));
                            task->setPriority(Normal);
                            dS_Assistant::getInstance()->addLongTask(task);
                        }
                        break;
                    }
                    case dsmsg::DEVICE_SCENE_TABLE_UPDATE_SUSPEND:
                    {
                        boost::shared_ptr<const dSLongTaskInfo> task = dS_Assistant::getInstance()->getTaskInfoByType(
                                TestDeviceUpdateTask);
                        if (task != NULL)
                        {
                            try
                            {
                                FILE_LOG(logINFO) << "[CommChannel] " <<
                                    "suspending task " << task->getName();
                                dS_Assistant::getInstance()->suspendTask(task->getTaskType());
                            }
                            catch (Exception &ex)
                            {
                                FILE_LOG(logWARNING) << "[CommChannel] " << ex.getMessage();
                            }
                        }
                        else
                        {
                            FILE_LOG(logWARNING) << "[CommChannel] No Task wit ID TestDeviceUpdateTask found";
                        }
                        break;
                    }
                    case dsmsg::DEVICE_SCENE_TABLE_UPDATE_ABORT:
                    {
                        boost::shared_ptr<const dSLongTaskInfo> task = dS_Assistant::getInstance()->getTaskInfoByType(
                                TestDeviceUpdateTask);
                        if (task != NULL)
                        {
                            try
                            {
                                dS_Assistant::getInstance()->cancelTask(task->getTaskID());
                            }
                            catch (Exception &ex)
                            {
                                FILE_LOG(logWARNING) << "[CommChannel] " << ex.getMessage();
                            }
                        }
                        else
                        {
                            FILE_LOG(logWARNING) << "[CommChannel] No Task wit ID TestDeviceUpdateTask found";
                        }
                        break;
                        break;
                    }
                    case dsmsg::DEVICE_SCENE_TABLE_UPDATE_RESUME:
                    {
                        boost::shared_ptr<const dSLongTaskInfo> task = dS_Assistant::getInstance()->getTaskInfoByType(
                                TestDeviceUpdateTask);
                        if (task != NULL)
                        {
                            try
                            {
                                FILE_LOG(logINFO) << "[CommChannel] " <<
                                    "resuming task " << task->getName();

                                dS_Assistant::getInstance()->resumeTask(task->getTaskType());
                            }
                            catch (Exception &ex)
                            {
                                FILE_LOG(logWARNING) << "[CommChannel] " << ex.getMessage();
                            }
                        }
                        else
                        {
                            FILE_LOG(logWARNING) << "[CommChannel] No Task wit ID TestDeviceUpdateTask found";
                        }
                        break;
                    }
                    case dsmsg::DEVICE_SCENE_TABLE_UPDATE_STATE_REQUEST:
                    {
                        boost::shared_ptr<const dSLongTaskInfo> task = dS_Assistant::getInstance()->getTaskInfoByType(
                                TestDeviceUpdateTask);

                        if (task == NULL)
                        {
                            FILE_LOG(logWARNING) << "[CommChannel] No Task wit ID TestDeviceUpdateTask found";
                            break;
                        }
                        if (const DeviceUpdateTask *pTask = dynamic_cast<const DeviceUpdateTask *>(task.get()))
                        {
                            const std::vector<int> sceneList = pTask->getSceneChanges();
                            response.set_type(dsmsg::DEVICE_SCENE_TABLE_UPDATE_STATE_RESPONSE);
                            dsmsg::DeviceSceneTableUpdateStateRes *resp = response.MutableExtension(
                                    dsmsg::device_scene_table_update_state_res);
                            resp->set_result(dsmsg::RESULT_OK);
                            for (std::vector<int>::const_iterator sceneIt = sceneList.begin();
                                    sceneIt != sceneList.end(); ++sceneIt)
                            {
                                resp->add_locked_scene(*sceneIt);
                            }
                        }
                        break;
                    }
                    default:
                    {
                        result = dsmsg::RESULT_MESSAGE_UNKNOWN;
                        resultDescription = "Message unknown";
                        break;
                    }
                }

                if (response.type() == dsmsg::GENERIC_RESPONSE)
                {
                    dsmsg::GenericResponse *resp = response.mutable_generic_res();
                    resp->set_result(result);
                    resp->set_description(resultDescription);
                }
                std::string responseString = response.SerializeAsString();

                try
                {
                    m_pServer->sendReply(c->getID(), (unsigned char *) responseString.c_str(), responseString.length());
                }
                catch (CC::Exception &ex)
                {
                    FILE_LOG(logERROR)
                                      << "[CommChannel] Could not send reply to "
                                      << util::str(c->getID().fd)
                                      << ": "
                                      << ex.getMessage();
                }
            }
            if (!m_shutdown_flag)
            {
                lockMessageList();
            }
        }
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << "[CommChannel] exit comm channel loop: " << ex.getMessage();
    }
    return;
}

}  // namespace dSA
